#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from neo4j.packstream import (
    UnpackableBuffer,
    Unpacker,
)

import logging
log = logging.getLogger("neo4j")


class MessageInbox:

    def __init__(self, s, on_error):
        self.on_error = on_error
        self._messages = self._yield_messages(s)

    def _yield_messages(self, sock):
        try:
            buffer = UnpackableBuffer()
            unpacker = Unpacker(buffer)
            chunk_size = 0
            while True:

                while chunk_size == 0:
                    # Determine the chunk size and skip noop
                    buffer.receive(sock, 2)
                    chunk_size = buffer.pop_u16()
                    if chunk_size == 0:
                        log.debug("[#%04X]  S: <NOOP>", sock.getsockname()[1])

                buffer.receive(sock, chunk_size + 2)
                chunk_size = buffer.pop_u16()

                if chunk_size == 0:
                    # chunk_size was the end marker for the message
                    size, tag = unpacker.unpack_structure_header()
                    fields = [unpacker.unpack() for _ in range(size)]
                    yield tag, fields
                    # Reset for new message
                    unpacker.reset()

        except OSError as error:
            self.on_error(error)

    def pop(self):
        return next(self._messages)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import subprocess
import os
import time

from threading import Thread
from time import sleep

from boltkit.server.stub import BoltStubService
from pytest import fixture

import logging
log = logging.getLogger("neo4j")

# from neo4j.debug import watch
# watch("neo4j")


class StubServer:

    def __init__(self, port, script):
        self.port = port
        self.script = os.path.join(os.path.dirname(__file__), "scripts", script)

    def run(self):
        self._process = subprocess.Popen(["python", "-m", "boltkit", "stub", "-v", "-l", ":{}".format(str(self.port)), "-t", "10", self.script], stdout=subprocess.PIPE)
        # Need verbose for this to work
        line = self._process.stdout.readline().decode("utf-8")
        log.debug("started stub server {}".format(self.port))
        log.debug(line.strip("\n"))

    def wait(self):
        while True:
            return_code = self._process.poll()
            if return_code is not None:
                line = self._process.stdout.readline().decode("utf-8")
                if line == "":
                    break
                log.debug(line.strip("\n"))

        return True

    def kill(self):
        # Kill process if not already dead
        if self._process.poll() is None:
            self._process.kill()


class StubCluster:

    def __init__(self, servers):
        self.servers = {port: StubServer(port, script) for port, script in dict(servers).items()}

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait()

    def start(self):
        for port, server in self.servers.items():
            server.run()

    def wait(self):
        success = True
        for port, server in self.servers.items():
            if not server.wait():
                success = False
            server.kill()

        if not success:
            raise Exception("Stub server failed")


class LegacyStubServer(Thread):

    def __init__(self, port, script):
        super(LegacyStubServer, self).__init__()
        self.port = port
        self.script = os.path.join(os.path.dirname(__file__), "scripts", script)

    def run(self):
        check_call(["python", "-m", "boltkit.legacy.stub", "-v", str(self.port), self.script])


class LegacyStubCluster:

    def __init__(self, servers):
        self.servers = {port: LegacyStubServer(port, script) for port, script in dict(servers).items()}

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait()

    def start(self):
        for port, server in self.servers.items():
            server.start()
        sleep(0.5)

    def wait(self):
        for port, server in self.servers.items():
            server.join()


class DefaultBoltStubService(BoltStubService):

    default_base_port = 9001


class StubCluster(StubCluster):

    def __init__(self, *servers):
        print("")
        scripts = [os.path.join(os.path.dirname(__file__), "scripts", server) for server in servers]

        bss = DefaultBoltStubService.load(*scripts)
        servers2 = {port: script.filename for port, script in bss.scripts.items()}
        super().__init__(servers2)

        # def run():
        #     check_call(["bolt", "stub", "-v", "-t", "10", "-l", ":9001"] + scripts)

        # self.thread = Thread(target=run)

    # def __enter__(self):
    #     self.thread.start()
    #     sleep(0.5)

    # def __exit__(self, exc_type, exc_value, traceback):
    #     self.thread.join(3)


@fixture
def script():
    return lambda *paths: path_join(dirname(__file__), "scripts", *paths)


@fixture
def driver_info():
    """ Base class for test cases that integrate with a server.
    """
    return {
        "uri_bolt": "bolt://localhost:9001",
        "uri_neo4j": "neo4j://localhost:9001",
        "user": "test",
        "password": "test",
        "auth_token": ("test", "test")
    }



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from math import ceil
from os import getenv
from os.path import dirname, join
from threading import RLock

import pytest
import urllib

from neo4j import (
    GraphDatabase,
)
from neo4j.exceptions import ServiceUnavailable
from neo4j._exceptions import BoltHandshakeError
from neo4j.io import Bolt

# import logging
# log = logging.getLogger("neo4j")
#
# from neo4j.debug import watch
# watch("neo4j")

NEO4J_RELEASES = getenv("NEO4J_RELEASES", "snapshot-enterprise 3.5-enterprise").split()
NEO4J_HOST = "localhost"
NEO4J_PORTS = {
    "bolt": 17601,
    "http": 17401,
    "https": 17301,
}
NEO4J_CORES = 3
NEO4J_REPLICAS = 2
NEO4J_USER = "neo4j"
NEO4J_PASSWORD = "password"
NEO4J_AUTH = (NEO4J_USER, NEO4J_PASSWORD)
NEO4J_LOCK = RLock()
NEO4J_SERVICE = None
NEO4J_DEBUG = getenv("NEO4J_DEBUG", "")

# TODO: re-enable when Docker is feasible
# from boltkit.server import Neo4jService, Neo4jClusterService


class Machine(object):

    def __init__(self, address):
        self.address = address


class Neo4jService(object):

    run = d = join(dirname(__file__), ".run")

    edition = "enterprise"

    def __init__(self, name=None, image=None, auth=None,
                 n_cores=None, n_replicas=None,
                 bolt_port=None, http_port=None, debug_port=None,
                 debug_suspend=None, dir_spec=None, config=None):
        from boltkit.legacy.controller import _install, create_controller
        assert image.endswith("-enterprise")
        release = image[:-11]
        if release == "snapshot":
            release = "4.0"
        self.home = _install("enterprise", release, self.run, verbose=1)
        self.auth = NEO4J_AUTH
        self.controller = create_controller(self.home)
        self.controller.set_initial_password(NEO4J_PASSWORD)
        self.info = None

    def start(self, timeout=None):
        self.info = self.controller.start(timeout=timeout)

    def stop(self, timeout=None):
        from shutil import rmtree
        self.controller.stop()
        rmtree(self.home)

    def machines(self):
        parsed = self.info.bolt_uri
        return [Machine((parsed.hostname, parsed.port or 7687))]

    @property
    def addresses(self):
        return [machine.address for machine in self.machines()]

    def cores(self):
        return self.machines()


class Neo4jClusterService(Neo4jService):

    @classmethod
    def _port_range(cls, base_port, count):
        if base_port is None:
            return [None] * count
        else:
            return range(base_port, base_port + count)


def _ping_range(port_range):
    count = 0
    for port in port_range:
        count += 1 if Bolt.ping((NEO4J_HOST, port)) else 0
    return count


def get_existing_service():
    core_port_range = Neo4jClusterService._port_range(NEO4J_PORTS["bolt"], NEO4J_CORES)
    replica_port_range = Neo4jClusterService._port_range(ceil(core_port_range.stop / 10) * 10 + 1,
                                                         NEO4J_REPLICAS)
    core_count = _ping_range(core_port_range)
    replica_count = _ping_range(replica_port_range)
    if core_count == 0 and replica_count == 0:
        return None
    elif core_count == NEO4J_CORES and replica_count == NEO4J_REPLICAS:
        return ExistingService(NEO4J_HOST, core_port_range, replica_port_range)
    else:
        raise OSError("Some ports required by the test service are already in use")


class ExistingMachine:

    def __init__(self, address):
        self.address = address


class ExistingService:

    def __init__(self, host, core_port_range, replica_port_range):
        self._cores = [ExistingMachine((host, port)) for port in core_port_range]
        self._replicas = [ExistingMachine((host, port)) for port in replica_port_range]

    @property
    def addresses(self):
        return [machine.address for machine in self.cores()]

    @property
    def auth(self):
        return NEO4J_AUTH

    def cores(self):
        return self._cores

    def replicas(self):
        return self._replicas

    def start(self, timeout=None):
        pass

    def stop(self, timeout=None):
        pass


existing_service = get_existing_service()

if existing_service:
    NEO4J_RELEASES = ["existing"]


@pytest.fixture(scope="session", params=NEO4J_RELEASES)
def service(request):
    global NEO4J_SERVICE
    if NEO4J_DEBUG:
        from neo4j.debug import watch
        watch("neo4j", "boltkit")
    with NEO4J_LOCK:
        assert NEO4J_SERVICE is None
        NEO4J_SERVICE = existing_service
        if existing_service:
            NEO4J_SERVICE = existing_service
        else:
            try:
                NEO4J_SERVICE = Neo4jService(auth=NEO4J_AUTH, image=request.param, n_cores=NEO4J_CORES, n_replicas=NEO4J_REPLICAS)
                NEO4J_SERVICE.start(timeout=300)
            except urllib.error.HTTPError as error:
                # pytest.skip(str(error))
                pytest.xfail(str(error) + " " + request.param)
        yield NEO4J_SERVICE
        if NEO4J_SERVICE is not None:
            NEO4J_SERVICE.stop(timeout=300)
            NEO4J_SERVICE = None


@pytest.fixture(scope="session")
def addresses(service):
    try:
        machines = service.cores()
    except AttributeError:
        machines = list(service.machines.values())
    return [machine.address for machine in machines]


# @fixture(scope="session")
# def readonly_addresses(service):
#     try:
#         machines = service.replicas()
#     except AttributeError:
#         machines = []
#     return [machine.address for machine in machines]


@pytest.fixture(scope="session")
def address(addresses):
    try:
        return addresses[0]
    except IndexError:
        return None


# @fixture(scope="session")
# def readonly_address(readonly_addresses):
#     try:
#         return readonly_addresses[0]
#     except IndexError:
#         return None


@pytest.fixture(scope="session")
def targets(addresses):
    return " ".join("{}:{}".format(address[0], address[1]) for address in addresses)


# @fixture(scope="session")
# def readonly_targets(addresses):
#     return " ".join("{}:{}".format(address[0], address[1]) for address in readonly_addresses)


@pytest.fixture(scope="session")
def target(address):
    return "{}:{}".format(address[0], address[1])


# @fixture(scope="session")
# def readonly_target(readonly_address):
#     if readonly_address:
#         return "{}:{}".format(readonly_address[0], readonly_address[1])
#     else:
#         return None


@pytest.fixture(scope="session")
def bolt_uri(service, target):
    return "bolt://" + target


@pytest.fixture(scope="session")
def neo4j_uri(service, target):
    return "neo4j://" + target


@pytest.fixture(scope="session")
def uri(bolt_uri):
    return bolt_uri


# @fixture(scope="session")
# def readonly_bolt_uri(service, readonly_target):
#     if readonly_target:
#         return "bolt://" + readonly_target
#     else:
#         return None


@pytest.fixture(scope="session")
def auth():
    return NEO4J_AUTH


@pytest.fixture(scope="session")
def bolt_driver(target, auth):
    try:
        driver = GraphDatabase.bolt_driver(target, auth=auth)
        try:
            yield driver
        finally:
            driver.close()
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])


@pytest.fixture(scope="session")
def neo4j_driver(target, auth):
    try:
        driver = GraphDatabase.neo4j_driver(target, auth=auth)
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])
        elif error.args[0] == "Server does not support routing":
            pytest.skip(error.args[0])
        else:
            raise
    else:
        try:
            yield driver
        finally:
            driver.close()


@pytest.fixture(scope="session")
def driver(neo4j_driver):
    return neo4j_driver


@pytest.fixture()
def session(bolt_driver):
    session = bolt_driver.session()
    try:
        yield session
    finally:
        session.close()


@pytest.fixture()
def protocol_version(session):
    result = session.run("RETURN 1")
    yield session._connection.server_info.protocol_version
    result.consume()


@pytest.fixture()
def cypher_eval(bolt_driver):

    def run_and_rollback(tx, cypher, **parameters):
        result = tx.run(cypher, **parameters)
        value = result.single().value()
        tx._success = False  # This is not a recommended pattern
        return value

    def f(cypher, **parameters):
        with bolt_driver.session() as session:
            return session.write_transaction(run_and_rollback, cypher, **parameters)

    return f


def pytest_sessionfinish(session, exitstatus):
    """ Called after the entire session to ensure Neo4j is shut down.
    """
    global NEO4J_SERVICE
    with NEO4J_LOCK:
        if NEO4J_SERVICE is not None:
            NEO4J_SERVICE.stop(timeout=300)
            NEO4J_SERVICE = None



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.io._courier import MessageInbox


class FakeSocket:

    def __init__(self, address):
        self.address = address
        self.captured = b""
        self.messages = MessageInbox(self, on_error=print)

    def getsockname(self):
        return "127.0.0.1", 0xFFFF

    def getpeername(self):
        return self.address

    def recv_into(self, buffer, nbytes):
        data = self.captured[:nbytes]
        actual = len(data)
        buffer[:actual] = data
        self.captured = self.captured[actual:]
        return actual

    def sendall(self, data):
        self.captured += data

    def close(self):
        return

    def pop_message(self):
        return self.messages.pop()


@pytest.fixture
def fake_socket():
    return FakeSocket



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABCMeta
from collections.abc import Mapping
from warnings import warn

from neo4j.meta import get_user_agent

from neo4j.api import (
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    TRUST_ALL_CERTIFICATES,
    WRITE_ACCESS,
    DEFAULT_DATABASE,
)
from neo4j.exceptions import (
    ConfigurationError,
)


def iter_items(iterable):
    """ Iterate through all items (key-value pairs) within an iterable
    dictionary-like object. If the object has a `keys` method, this is
    used along with `__getitem__` to yield each pair in turn. If no
    `keys` method exists, each iterable element is assumed to be a
    2-tuple of key and value.
    """
    if hasattr(iterable, "keys"):
        for key in iterable.keys():
            yield key, iterable[key]
    else:
        for key, value in iterable:
            yield key, value


class DeprecatedAlias:

    def __init__(self, new):
        self.new = new


class ConfigType(ABCMeta):

    def __new__(mcs, name, bases, attributes):
        fields = []
        deprecated_aliases = {}

        for base in bases:
            if type(base) is mcs:
                fields += base.keys()
                deprecated_aliases.update(base._deprecated_aliases())

        for k, v in attributes.items():
            if isinstance(v, DeprecatedAlias):
                deprecated_aliases[k] = v.new
            elif not k.startswith("_") and not callable(v):
                fields.append(k)

        def keys(_):
            return fields

        def _deprecated_aliases(_):
            return deprecated_aliases

        def _deprecated_keys(_):
            return list(deprecated_aliases)

        def _get_new(_, key):
            return deprecated_aliases.get(key)

        attributes.setdefault("keys", classmethod(keys))
        attributes.setdefault("_deprecated_aliases", classmethod(_deprecated_aliases))
        attributes.setdefault("_deprecated_keys", classmethod(_deprecated_keys))
        attributes.setdefault("_get_new", classmethod(_get_new))

        return super(ConfigType, mcs).__new__(mcs, name, bases,
                                              {k: v for k, v in attributes.items()
                                               if k not in deprecated_aliases})


class Config(Mapping, metaclass=ConfigType):
    """ Base class for all configuration containers.
    """

    @staticmethod
    def consume_chain(data, *config_classes):
        values = []
        for config_class in config_classes:
            if not issubclass(config_class, Config):
                raise TypeError("%r is not a Config subclass" % config_class)
            values.append(config_class._consume(data))
        if data:
            raise ConfigurationError("Unexpected config keys: %s" % ", ".join(data.keys()))
        return values

    @classmethod
    def consume(cls, data):
        config, = cls.consume_chain(data, cls)
        return config

    @classmethod
    def _consume(cls, data):
        config = {}
        if data:
            for key in list(cls.keys()) + list(cls._deprecated_keys()):
                try:
                    value = data.pop(key)
                except KeyError:
                    pass
                else:
                    config[key] = value
        return cls(config)

    def __update(self, data):
        data_dict = dict(iter_items(data))

        def set_attr(k, v):
            if k in self.keys():
                setattr(self, k, v)
            elif k in self._deprecated_keys():
                k0 = self._get_new(k)
                if k0 in data_dict:
                    raise ValueError("Cannot specify both '{}' and '{}' in config".format(k0, k))
                warn("The '{}' config key is deprecated, please use '{}' instead".format(k, k0))
                set_attr(k0, v)
            else:
                raise AttributeError(k)

        for key, value in data_dict.items():
            if value is not None:
                set_attr(key, value)

    def __init__(self, *args, **kwargs):
        for arg in args:
            self.__update(arg)
        self.__update(kwargs)

    def __repr__(self):
        attrs = []
        for key in self:
            attrs.append(" %s=%r" % (key, getattr(self, key)))
        return "<%s%s>" % (self.__class__.__name__, "".join(attrs))

    def __len__(self):
        return len(self.keys())

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        return iter(self.keys())


class PoolConfig(Config):
    """ Connection pool configuration.
    """

    #: Max Connection Lifetime
    max_connection_lifetime = 3600  # seconds
    # The maximum duration the driver will keep a connection for before being removed from the pool.

    #: Max Connection Pool Size
    max_connection_pool_size = 100
    # The maximum total number of connections allowed, per host (i.e. cluster nodes), to be managed by the connection pool.

    #: Connection Timeout
    connection_timeout = 30.0  # seconds
    # The maximum amount of time to wait for a TCP connection to be established.

    #: Trust
    trust = TRUST_SYSTEM_CA_SIGNED_CERTIFICATES
    # Specify how to determine the authenticity of encryption certificates provided by the Neo4j instance on connection.

    #: Custom Resolver
    resolver = None
    # Custom resolver function, returning list of resolved addresses.

    #: Encrypted
    encrypted = False
    # Specify whether to use an encrypted connection between the driver and server.

    #: User Agent (Python Driver Specific)
    user_agent = get_user_agent()
    # Specify the client agent name.

    #: Protocol Version (Python Driver Specific)
    protocol_version = None  # Version(4, 0)
    # Specify a specific Bolt Protocol Version

    #: Initial Connection Pool Size (Python Driver Specific)
    init_size = 1  # The other drivers do not seed from the start.
    # This will seed the pool with the specified number of connections.

    #: Socket Keep Alive (Python and .NET Driver Specific)
    keep_alive = True
    # Specify whether TCP keep-alive should be enabled.

    def get_ssl_context(self):
        if not self.encrypted:
            return None

        import ssl

        ssl_context = None

        # SSL stands for Secure Sockets Layer and was originally created by Netscape.
        # SSLv2 and SSLv3 are the 2 versions of this protocol (SSLv1 was never publicly released).
        # After SSLv3, SSL was renamed to TLS.
        # TLS stands for Transport Layer Security and started with TLSv1.0 which is an upgraded version of SSLv3.

        # SSLv2 - (Disabled)
        # SSLv3 - (Disabled)
        # TLS 1.0 - Released in 1999, published as RFC 2246. (Disabled)
        # TLS 1.1 - Released in 2006, published as RFC 4346. (Disabled)
        # TLS 1.2 - Released in 2008, published as RFC 5246.

        try:
            # python 3.6+
            # https://docs.python.org/3.6/library/ssl.html#ssl.PROTOCOL_TLS_CLIENT
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

            # For recommended security options see
            # https://docs.python.org/3.6/library/ssl.html#protocol-versions
            ssl_context.options |= ssl.OP_NO_TLSv1      # Python 3.2
            ssl_context.options |= ssl.OP_NO_TLSv1_1    # Python 3.4

        except AttributeError:
            # python 3.5
            # https://docs.python.org/3.5/library/ssl.html#ssl.PROTOCOL_TLS
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)

            # For recommended security options see
            # https://docs.python.org/3.5/library/ssl.html#protocol-versions
            ssl_context.options |= ssl.OP_NO_SSLv2      # Python 3.2
            ssl_context.options |= ssl.OP_NO_SSLv3      # Python 3.2
            ssl_context.options |= ssl.OP_NO_TLSv1      # Python 3.2
            ssl_context.options |= ssl.OP_NO_TLSv1_1    # Python 3.4

            ssl_context.verify_mode = ssl.CERT_REQUIRED     # https://docs.python.org/3.5/library/ssl.html#ssl.SSLContext.verify_mode
            ssl_context.check_hostname = True               # https://docs.python.org/3.5/library/ssl.html#ssl.SSLContext.check_hostname

        if self.trust == TRUST_ALL_CERTIFICATES:
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE     # https://docs.python.org/3.5/library/ssl.html#ssl.CERT_NONE

        ssl_context.set_default_verify_paths()  # https://docs.python.org/3.5/library/ssl.html#ssl.SSLContext.set_default_verify_paths
        return ssl_context


class WorkspaceConfig(Config):
    """ WorkSpace configuration.
    """

    #: Connection Acquisition Timeout
    connection_acquisition_timeout = 60.0  # seconds
    # The maximum amount of time a session will wait when requesting a connection from the connection pool.
    # Since the process of acquiring a connection may involve creating a new connection, ensure that the value
    # of this configuration is higher than the configured Connection Timeout.

    #: Max Transaction Retry Time
    max_transaction_retry_time = 30.0  # seconds
    # The maximum amount of time that a managed transaction will retry before failing.

    #: Initial Retry Delay
    initial_retry_delay = 1.0  # seconds

    #: Retry Delay Multiplier
    retry_delay_multiplier = 2.0  # seconds

    #: Retry Delay Jitter Factor
    retry_delay_jitter_factor = 0.2  # seconds

    #: Database Name
    database = DEFAULT_DATABASE
    # Name of the database to query.
    # Note: The default database can be set on the Neo4j instance settings.

    #: Fetch Size
    fetch_size = 1000


class SessionConfig(WorkspaceConfig):
    """ Session configuration.
    """

    #: Bookmarks
    bookmarks = ()

    #: Default AccessMode
    default_access_mode = WRITE_ACCESS
    # access_mode = DeprecatedAlias("default_access_mode")


class TransactionConfig(Config):
    """ Transaction configuration. This is internal for now.

    neo4j.session.begin_transaction
    neo4j.Query
    neo4j.unit_of_work

    are both using the same settings.
    """
    #: Metadata
    metadata = None  # dictionary

    #: Timeout
    timeout = None  # seconds


class RoutingConfig(Config):
    """ Neo4jDriver routing settings. This is internal for now.
    """

    #: Routing Table Purge_Delay
    routing_table_purge_delay = 30.0  # seconds
    # The TTL + routing_table_purge_delay should be used to check if the database routing table should be removed.

    #: Max Routing Failures
    # max_routing_failures = 1

    #: Retry Timeout Delay
    # retry_timeout_delay = 5.0  # seconds



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "__version__",
    "GraphDatabase",
    "Driver",
    "BoltDriver",
    "Neo4jDriver",
    "Auth",
    "AuthToken",
    "basic_auth",
    "kerberos_auth",
    "custom_auth",
    "Bookmark",
    "ServerInfo",
    "Version",
    "READ_ACCESS",
    "WRITE_ACCESS",
    "DEFAULT_DATABASE",
    "TRUST_ALL_CERTIFICATES",
    "TRUST_SYSTEM_CA_SIGNED_CERTIFICATES",
    "Address",
    "IPv4Address",
    "IPv6Address",
    "Config",
    "PoolConfig",
    "WorkspaceConfig",
    "SessionConfig",
    "Record",
    "Transaction",
    "Result",
    "ResultSummary",
    "SummaryCounters",
    "Query",
    "Session",
    "unit_of_work",
    "ExperimentalWarning",
]

from logging import getLogger


from neo4j.addressing import (
    Address,
    IPv4Address,
    IPv6Address,
)
from neo4j.api import (
    Auth,  # TODO: Validate naming for Auth compared to other drivers.
    AuthToken,
    basic_auth,
    kerberos_auth,
    custom_auth,
    Bookmark,
    ServerInfo,
    Version,
    READ_ACCESS,
    WRITE_ACCESS,
    SYSTEM_DATABASE,
    DEFAULT_DATABASE,
    TRUST_ALL_CERTIFICATES,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
)
from neo4j.conf import (
    Config,
    PoolConfig,
    WorkspaceConfig,
    SessionConfig,
)
from neo4j.meta import (
    experimental,
    ExperimentalWarning,
    get_user_agent,
    version as __version__,
)
from neo4j.data import (
    Record,
)
from neo4j.work.simple import (
    Query,
    Session,
    unit_of_work,
)
from neo4j.work.transaction import (
    Transaction,
)
from neo4j.work.result import (
    Result,
)
from neo4j.work.summary import (
    ResultSummary,
    SummaryCounters,
)


log = getLogger("neo4j")


class GraphDatabase:
    """Accessor for :class:`neo4j.Driver` construction.
    """

    @classmethod
    def driver(cls, uri, *, auth=None, **config):
        """Create a driver.

        :param uri: the connection URI for the driver, see :ref:`uri-ref` for available URIs.
        :param auth: the authentication details, see :ref:`auth-ref` for available authentication details.
        :param config: driver configuration key-word arguments, see :ref:`driver-configuration-ref` for available key-word arguments.

        :return: :ref:`neo4j-driver-ref` or :ref:`bolt-driver-ref`
        """

        from neo4j.api import (
            parse_neo4j_uri,
            parse_routing_context,
            DRIVER_BOLT,
            DRIVER_NEO4j,
            SECURITY_TYPE_NOT_SECURE,
            SECURITY_TYPE_SELF_SIGNED_CERTIFICATE,
            SECURITY_TYPE_SECURE,
            URI_SCHEME_BOLT,
            URI_SCHEME_NEO4J,
            URI_SCHEME_BOLT_SELF_SIGNED_CERTIFICATE,
            URI_SCHEME_BOLT_SECURE,
            URI_SCHEME_NEO4J_SELF_SIGNED_CERTIFICATE,
            URI_SCHEME_NEO4J_SECURE,
        )

        driver_type, security_type, parsed = parse_neo4j_uri(uri)

        if "trust" in config.keys():
            if config.get("trust") not in [TRUST_ALL_CERTIFICATES, TRUST_SYSTEM_CA_SIGNED_CERTIFICATES]:
                from neo4j.exceptions import ConfigurationError
                raise ConfigurationError("The config setting `trust` values are {!r}".format(
                    [
                        TRUST_ALL_CERTIFICATES,
                        TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
                    ]
                ))

        if security_type in [SECURITY_TYPE_SELF_SIGNED_CERTIFICATE, SECURITY_TYPE_SECURE] and ("encrypted" in config.keys() or "trust" in config.keys()):
            from neo4j.exceptions import ConfigurationError
            raise ConfigurationError("The config settings 'encrypted' and 'trust' can only be used with the URI schemes {!r}. Use the other URI schemes {!r} for setting encryption settings.".format(
                [
                    URI_SCHEME_BOLT,
                    URI_SCHEME_NEO4J,
                ],
                [
                    URI_SCHEME_BOLT_SELF_SIGNED_CERTIFICATE,
                    URI_SCHEME_BOLT_SECURE,
                    URI_SCHEME_NEO4J_SELF_SIGNED_CERTIFICATE,
                    URI_SCHEME_NEO4J_SECURE,
                ]
            ))

        if security_type == SECURITY_TYPE_SECURE:
            config["encrypted"] = True
        elif security_type == SECURITY_TYPE_SELF_SIGNED_CERTIFICATE:
            config["encrypted"] = True
            config["trust"] = TRUST_ALL_CERTIFICATES

        if driver_type == DRIVER_BOLT:
            return cls.bolt_driver(parsed.netloc, auth=auth, **config)
        elif driver_type == DRIVER_NEO4j:
            routing_context = parse_routing_context(parsed.query)
            return cls.neo4j_driver(parsed.netloc, auth=auth, routing_context=routing_context, **config)

    @classmethod
    def bolt_driver(cls, target, *, auth=None, **config):
        """ Create a driver for direct Bolt server access that uses
        socket I/O and thread-based concurrency.
        """
        from neo4j._exceptions import BoltHandshakeError, BoltSecurityError

        try:
            return BoltDriver.open(target, auth=auth, **config)
        except (BoltHandshakeError, BoltSecurityError) as error:
            from neo4j.exceptions import ServiceUnavailable
            raise ServiceUnavailable(str(error)) from error

    @classmethod
    def neo4j_driver(cls, *targets, auth=None, routing_context=None, **config):
        """ Create a driver for routing-capable Neo4j service access
        that uses socket I/O and thread-based concurrency.
        """
        from neo4j._exceptions import BoltHandshakeError, BoltSecurityError

        try:
            return Neo4jDriver.open(*targets, auth=auth, routing_context=routing_context, **config)
        except (BoltHandshakeError, BoltSecurityError) as error:
            from neo4j.exceptions import ServiceUnavailable
            raise ServiceUnavailable(str(error)) from error


class Direct:

    default_host = "localhost"
    default_port = 7687

    default_target = ":"

    def __init__(self, address):
        self._address = address

    @property
    def address(self):
        return self._address

    @classmethod
    def parse_target(cls, target):
        """ Parse a target string to produce an address.
        """
        if not target:
            target = cls.default_target
        address = Address.parse(target, default_host=cls.default_host,
                                default_port=cls.default_port)
        return address


class Routing:

    default_host = "localhost"
    default_port = 7687

    default_targets = ": :17601 :17687"

    def __init__(self, initial_addresses):
        self._initial_addresses = initial_addresses

    @property
    def initial_addresses(self):
        return self._initial_addresses

    @classmethod
    def parse_targets(cls, *targets):
        """ Parse a sequence of target strings to produce an address
        list.
        """
        targets = " ".join(targets)
        if not targets:
            targets = cls.default_targets
        addresses = Address.parse_list(targets, default_host=cls.default_host, default_port=cls.default_port)
        return addresses


class Driver:
    """ Base class for all types of :class:`neo4j.Driver`, instances of which are
    used as the primary access point to Neo4j.
    """

    #: Connection pool
    _pool = None

    def __init__(self, pool):
        assert pool is not None
        self._pool = pool

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def encrypted(self):
        return bool(self._pool.pool_config.encrypted)

    def session(self, **config):
        """Create a session, see :ref:`session-construction-ref`

        :param config: session configuration key-word arguments, see :ref:`session-configuration-ref` for available key-word arguments.

        :returns: new :class:`neo4j.Session` object
        """
        raise NotImplementedError

    @experimental("The pipeline API is experimental and may be removed or changed in a future release")
    def pipeline(self, **config):
        """ Create a pipeline.
        """
        raise NotImplementedError

    def close(self):
        """ Shut down, closing any open connections in the pool.
        """
        self._pool.close()

    @experimental("The configuration may change in the future.")
    def verify_connectivity(self, **config):
        """ This verifies if the driver can connect to a remote server or a cluster
        by establishing a network connection with the remote and possibly exchanging
        a few data before closing the connection. It throws exception if fails to connect.

        Use the exception to further understand the cause of the connectivity problem.

        Note: Even if this method throws an exception, the driver still need to be closed via close() to free up all resources.
        """
        raise NotImplementedError

    @experimental("Feature support query, based on Bolt Protocol Version and Neo4j Server Version will change in the future.")
    def supports_multi_db(self):
        """ Check if the server or cluster supports multi-databases.

        :return: Returns true if the server or cluster the driver connects to supports multi-databases, otherwise false.
        :rtype: bool
        """
        cx = self._pool.acquire(access_mode=READ_ACCESS, timeout=self._pool.workspace_config.connection_acquisition_timeout, database=self._pool.workspace_config.database)
        support = cx.supports_multiple_databases
        self._pool.release(cx)

        return support


class BoltDriver(Direct, Driver):
    """ A :class:`.BoltDriver` is created from a ``bolt`` URI and addresses
    a single database machine. This may be a standalone server or could be a
    specific member of a cluster.

    Connections established by a :class:`.BoltDriver` are always made to the
    exact host and port detailed in the URI.
    """

    @classmethod
    def open(cls, target, *, auth=None, **config):
        """
        :param target:
        :param auth:
        :param config: The values that can be specified are found in :class: `neo4j.PoolConfig` and :class: `neo4j.WorkspaceConfig`

        :return:
        :rtype: :class: `neo4j.BoltDriver`
        """
        from neo4j.io import BoltPool
        address = cls.parse_target(target)
        pool_config, default_workspace_config = Config.consume_chain(config, PoolConfig, WorkspaceConfig)
        pool = BoltPool.open(address, auth=auth, pool_config=pool_config, workspace_config=default_workspace_config)
        return cls(pool, default_workspace_config)

    def __init__(self, pool, default_workspace_config):
        Direct.__init__(self, pool.address)
        Driver.__init__(self, pool)
        self._default_workspace_config = default_workspace_config

    def session(self, **config):
        """
        :param config: The values that can be specified are found in :class: `neo4j.SessionConfig`

        :return:
        :rtype: :class: `neo4j.Session`
        """
        from neo4j.work.simple import Session
        session_config = SessionConfig(self._default_workspace_config, config)
        SessionConfig.consume(config)  # Consume the config
        return Session(self._pool, session_config)

    def pipeline(self, **config):
        from neo4j.work.pipelining import Pipeline, PipelineConfig
        pipeline_config = PipelineConfig(self._default_workspace_config, config)
        PipelineConfig.consume(config)  # Consume the config
        return Pipeline(self._pool, pipeline_config)

    @experimental("The configuration may change in the future.")
    def verify_connectivity(self, **config):
        server_agent = None
        config["fetch_size"] = -1
        with self.session(**config) as session:
            result = session.run("RETURN 1 AS x")
            value = result.single().value()
            summary = result.consume()
            server_agent = summary.server.agent
        return server_agent


class Neo4jDriver(Routing, Driver):
    """ A :class:`.Neo4jDriver` is created from a ``neo4j`` URI. The
    routing behaviour works in tandem with Neo4j's `Causal Clustering
    <https://neo4j.com/docs/operations-manual/current/clustering/>`_
    feature by directing read and write behaviour to appropriate
    cluster members.
    """

    @classmethod
    def open(cls, *targets, auth=None, routing_context=None, **config):
        from neo4j.io import Neo4jPool
        addresses = cls.parse_targets(*targets)
        pool_config, default_workspace_config = Config.consume_chain(config, PoolConfig, WorkspaceConfig)
        pool = Neo4jPool.open(*addresses, auth=auth, routing_context=routing_context, pool_config=pool_config, workspace_config=default_workspace_config)
        return cls(pool, default_workspace_config)

    def __init__(self, pool, default_workspace_config):
        Routing.__init__(self, pool.get_default_database_initial_router_addresses())
        Driver.__init__(self, pool)
        self._default_workspace_config = default_workspace_config

    def session(self, **config):
        session_config = SessionConfig(self._default_workspace_config, config)
        SessionConfig.consume(config)  # Consume the config
        return Session(self._pool, session_config)

    def pipeline(self, **config):
        from neo4j.work.pipelining import Pipeline, PipelineConfig
        pipeline_config = PipelineConfig(self._default_workspace_config, config)
        PipelineConfig.consume(config)  # Consume the config
        return Pipeline(self._pool, pipeline_config)

    @experimental("The configuration may change in the future.")
    def verify_connectivity(self, **config):
        """
        :raise ServiceUnavailable: raised if the server does not support routing or if routing support is broken.
        """
        # TODO: Improve and update Stub Test Server to be able to test.
        return self._verify_routing_connectivity()

    def _verify_routing_connectivity(self):
        from neo4j.exceptions import ServiceUnavailable
        from neo4j._exceptions import BoltHandshakeError

        table = self._pool.get_routing_table_for_default_database()
        routing_info = {}
        for ix in list(table.routers):
            try:
                routing_info[ix] = self._pool.fetch_routing_info(address=table.routers[0], timeout=self._default_workspace_config.connection_acquisition_timeout, database=self._default_workspace_config.database)
            except BoltHandshakeError as error:
                routing_info[ix] = None

        for key, val in routing_info.items():
            if val is not None:
                return routing_info
        raise ServiceUnavailable("Could not connect to any routing servers.")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.exceptions import (
    Neo4jError,
    ClientError,
    CypherSyntaxError,
    CypherTypeError,
    ConstraintError,
    AuthError,
    Forbidden,
    ForbiddenOnReadOnlyDatabase,
    NotALeader,
    DatabaseError,
    TransientError,
    DatabaseUnavailable,
    DriverError,
    TransactionError,
    TransactionNestingError,
    SessionExpired,
    ServiceUnavailable,
    RoutingServiceUnavailable,
    WriteServiceUnavailable,
    ReadServiceUnavailable,
    ConfigurationError,
    AuthConfigurationError,
    CertificateConfigurationError,
    ResultConsumedError,
    CLASSIFICATION_CLIENT,
    CLASSIFICATION_DATABASE,
    CLASSIFICATION_TRANSIENT,
)

from neo4j._exceptions import (
    BoltError,
    BoltHandshakeError,
    BoltRoutingError,
    BoltConnectionError,
    BoltSecurityError,
    BoltConnectionBroken,
    BoltConnectionClosed,
    BoltFailure,
    BoltIncompleteCommitError,
    BoltProtocolError,
)

from neo4j.io import Bolt


# python -m pytest tests/unit/test_exceptions.py -s -v

def test_bolt_error():
    with pytest.raises(BoltError) as e:
        error = BoltError("Error Message", address="localhost")
        # assert repr(error) == "BoltError('Error Message')" This differs between python version 3.6 "BoltError('Error Message',)"  and 3.7
        assert str(error) == "Error Message"
        assert error.args == ("Error Message",)
        assert error.address == "localhost"
        raise error

    # The regexp parameter of the match method is matched with the re.search function.
    with pytest.raises(AssertionError):
        e.match("FAIL!")

    assert e.match("Error Message")


def test_bolt_protocol_error():
    with pytest.raises(BoltProtocolError) as e:
        error = BoltProtocolError("Driver does not support Bolt protocol version: 0x%06X%02X" % (2, 5), address="localhost")
        assert error.address == "localhost"
        raise error

    # The regexp parameter of the match method is matched with the re.search function.
    with pytest.raises(AssertionError):
        e.match("FAIL!")

    e.match("Driver does not support Bolt protocol version: 0x00000205")


def test_bolt_handshake_error():
    handshake = b"\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00"
    response = b"\x00\x00\x00\x00"
    supported_versions = Bolt.protocol_handlers().keys()

    with pytest.raises(BoltHandshakeError) as e:
        error = BoltHandshakeError("The Neo4J server does not support communication with this driver. Supported Bolt Protocols {}".format(supported_versions), address="localhost", request_data=handshake, response_data=response)
        assert error.address == "localhost"
        assert error.request_data == handshake
        assert error.response_data == response
        raise error

    e.match("The Neo4J server does not support communication with this driver. Supported Bolt Protocols ")


def test_serviceunavailable():
    with pytest.raises(ServiceUnavailable) as e:
        error = ServiceUnavailable("Test error message")
        raise error

    assert e.value.__cause__ is None


def test_serviceunavailable_raised_from_bolt_protocol_error_with_implicit_style():
    error = BoltProtocolError("Driver does not support Bolt protocol version: 0x%06X%02X" % (2, 5), address="localhost")
    with pytest.raises(ServiceUnavailable) as e:
        assert error.address == "localhost"
        try:
            raise error
        except BoltProtocolError as error_bolt_protocol:
            raise ServiceUnavailable(str(error_bolt_protocol)) from error_bolt_protocol

    # The regexp parameter of the match method is matched with the re.search function.
    with pytest.raises(AssertionError):
        e.match("FAIL!")

    e.match("Driver does not support Bolt protocol version: 0x00000205")
    assert e.value.__cause__ is error


def test_serviceunavailable_raised_from_bolt_protocol_error_with_explicit_style():
    error = BoltProtocolError("Driver does not support Bolt protocol version: 0x%06X%02X" % (2, 5), address="localhost")

    with pytest.raises(ServiceUnavailable) as e:
        assert error.address == "localhost"
        try:
            raise error
        except BoltProtocolError as error_bolt_protocol:
            error_nested = ServiceUnavailable(str(error_bolt_protocol))
            error_nested.__cause__ = error_bolt_protocol
            raise error_nested

    # The regexp parameter of the match method is matched with the re.search function.
    with pytest.raises(AssertionError):
        e.match("FAIL!")

    e.match("Driver does not support Bolt protocol version: 0x00000205")
    assert e.value.__cause__ is error


def test_neo4jerror_hydrate_with_no_args():
    error = Neo4jError.hydrate()

    assert isinstance(error, DatabaseError)
    assert error.classification == CLASSIFICATION_DATABASE
    assert error.category == "General"
    assert error.title == "UnknownError"
    assert error.metadata == {}
    assert error.message == "An unknown error occurred"
    assert error.code == "Neo.DatabaseError.General.UnknownError"


def test_neo4jerror_hydrate_with_message_and_code_rubish():
    error = Neo4jError.hydrate(message="Test error message", code="ASDF_asdf")

    assert isinstance(error, DatabaseError)
    assert error.classification == CLASSIFICATION_DATABASE
    assert error.category == "General"
    assert error.title == "UnknownError"
    assert error.metadata == {}
    assert error.message == "Test error message"
    assert error.code == "ASDF_asdf"


def test_neo4jerror_hydrate_with_message_and_code_database():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.DatabaseError.General.UnknownError")

    assert isinstance(error, DatabaseError)
    assert error.classification == CLASSIFICATION_DATABASE
    assert error.category == "General"
    assert error.title == "UnknownError"
    assert error.metadata == {}
    assert error.message == "Test error message"
    assert error.code == "Neo.DatabaseError.General.UnknownError"


def test_neo4jerror_hydrate_with_message_and_code_transient():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.TransientError.General.TestError")
    # error = Neo4jError.hydrate(message="Test error message", code="Neo.{}.General.TestError".format(CLASSIFICATION_TRANSIENT))

    assert isinstance(error, TransientError)
    assert error.classification == CLASSIFICATION_TRANSIENT
    assert error.category == "General"
    assert error.title == "TestError"
    assert error.metadata == {}
    assert error.message == "Test error message"
    assert error.code == "Neo.{}.General.TestError".format(CLASSIFICATION_TRANSIENT)


def test_neo4jerror_hydrate_with_message_and_code_client():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.{}.General.TestError".format(CLASSIFICATION_CLIENT))

    assert isinstance(error, ClientError)
    assert error.classification == CLASSIFICATION_CLIENT
    assert error.category == "General"
    assert error.title == "TestError"
    assert error.metadata == {}
    assert error.message == "Test error message"
    assert error.code == "Neo.{}.General.TestError".format(CLASSIFICATION_CLIENT)


def test_transient_error_is_retriable_case_1():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.TransientError.Transaction.Terminated")

    assert isinstance(error, TransientError)
    assert error.is_retriable() is False


def test_transient_error_is_retriable_case_2():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.TransientError.Transaction.LockClientStopped")

    assert isinstance(error, TransientError)
    assert error.is_retriable() is False


def test_transient_error_is_retriable_case_3():
    error = Neo4jError.hydrate(message="Test error message", code="Neo.TransientError.General.TestError")

    assert isinstance(error, TransientError)
    assert error.is_retriable() is True



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import date
from time import struct_time
from unittest import TestCase

import pytz

from neo4j.time import Duration, Date, UnixEpoch, ZeroDate


eastern = pytz.timezone("US/Eastern")


class DateTestCase(TestCase):

    def test_bad_attribute(self):
        d = Date(2000, 1, 1)
        with self.assertRaises(AttributeError):
            _ = d.x

    def test_zero_date(self):
        d = Date(0, 0, 0)
        self.assertEqual(d.year_month_day, (0, 0, 0))
        self.assertEqual(d.year, 0)
        self.assertEqual(d.month, 0)
        self.assertEqual(d.day, 0)
        self.assertIs(d, ZeroDate)

    def test_zero_ordinal(self):
        d = Date.from_ordinal(0)
        self.assertEqual(d.year_month_day, (0, 0, 0))
        self.assertEqual(d.year, 0)
        self.assertEqual(d.month, 0)
        self.assertEqual(d.day, 0)
        self.assertIs(d, ZeroDate)

    def test_ordinal_at_start_of_1970(self):
        d = Date.from_ordinal(719163)
        self.assertEqual(d.year_month_day, (1970, 1, 1))
        self.assertEqual(d.year, 1970)
        self.assertEqual(d.month, 1)
        self.assertEqual(d.day, 1)

    def test_ordinal_at_end_of_1969(self):
        d = Date.from_ordinal(719162)
        self.assertEqual(d.year_month_day, (1969, 12, 31))
        self.assertEqual(d.year, 1969)
        self.assertEqual(d.month, 12)
        self.assertEqual(d.day, 31)

    def test_ordinal_at_start_of_2018(self):
        d = Date.from_ordinal(736695)
        self.assertEqual(d.year_month_day, (2018, 1, 1))
        self.assertEqual(d.year, 2018)
        self.assertEqual(d.month, 1)
        self.assertEqual(d.day, 1)

    def test_ordinal_at_end_of_2017(self):
        d = Date.from_ordinal(736694)
        self.assertEqual(d.year_month_day, (2017, 12, 31))
        self.assertEqual(d.year, 2017)
        self.assertEqual(d.month, 12)
        self.assertEqual(d.day, 31)

    def test_all_positive_days_of_month_for_31_day_month(self):
        for day in range(1, 32):
            t = Date(1976, 1, day)
            self.assertEqual(t.year_month_day, (1976, 1, day))
            self.assertEqual(t.year, 1976)
            self.assertEqual(t.month, 1)
            self.assertEqual(t.day, day)
        with self.assertRaises(ValueError):
            _ = Date(1976, 1, 32)

    def test_all_positive_days_of_month_for_30_day_month(self):
        for day in range(1, 31):
            t = Date(1976, 6, day)
            self.assertEqual(t.year_month_day, (1976, 6, day))
            self.assertEqual(t.year, 1976)
            self.assertEqual(t.month, 6)
            self.assertEqual(t.day, day)
        with self.assertRaises(ValueError):
            _ = Date(1976, 6, 31)

    def test_all_positive_days_of_month_for_29_day_month(self):
        for day in range(1, 30):
            t = Date(1976, 2, day)
            self.assertEqual(t.year_month_day, (1976, 2, day))
            self.assertEqual(t.year, 1976)
            self.assertEqual(t.month, 2)
            self.assertEqual(t.day, day)
        with self.assertRaises(ValueError):
            _ = Date(1976, 2, 30)

    def test_all_positive_days_of_month_for_28_day_month(self):
        for day in range(1, 29):
            t = Date(1977, 2, day)
            self.assertEqual(t.year_month_day, (1977, 2, day))
            self.assertEqual(t.year, 1977)
            self.assertEqual(t.month, 2)
            self.assertEqual(t.day, day)
        with self.assertRaises(ValueError):
            _ = Date(1977, 2, 29)

    def test_last_but_2_day_for_31_day_month(self):
        t = Date(1976, 1, -3)
        self.assertEqual(t.year_month_day, (1976, 1, 29))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 29)

    def test_last_but_1_day_for_31_day_month(self):
        t = Date(1976, 1, -2)
        self.assertEqual(t.year_month_day, (1976, 1, 30))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 30)

    def test_last_day_for_31_day_month(self):
        t = Date(1976, 1, -1)
        self.assertEqual(t.year_month_day, (1976, 1, 31))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 1)
        self.assertEqual(t.day, 31)

    def test_last_but_1_day_for_30_day_month(self):
        t = Date(1976, 6, -2)
        self.assertEqual(t.year_month_day, (1976, 6, 29))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 6)
        self.assertEqual(t.day, 29)

    def test_last_day_for_30_day_month(self):
        t = Date(1976, 6, -1)
        self.assertEqual(t.year_month_day, (1976, 6, 30))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 6)
        self.assertEqual(t.day, 30)

    def test_day_28_for_29_day_month(self):
        t = Date(1976, 2, 28)
        self.assertEqual(t.year_month_day, (1976, 2, 28))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 2)
        self.assertEqual(t.day, 28)

    def test_last_day_for_29_day_month(self):
        t = Date(1976, 2, -1)
        self.assertEqual(t.year_month_day, (1976, 2, 29))
        self.assertEqual(t.year, 1976)
        self.assertEqual(t.month, 2)
        self.assertEqual(t.day, 29)

    def test_last_day_for_28_day_month(self):
        t = Date(1977, 2, -1)
        self.assertEqual(t.year_month_day, (1977, 2, 28))
        self.assertEqual(t.year, 1977)
        self.assertEqual(t.month, 2)
        self.assertEqual(t.day, 28)

    def test_cannot_use_year_lower_than_one(self):
        with self.assertRaises(ValueError):
            _ = Date(0, 2, 1)

    def test_cannot_use_year_higher_than_9999(self):
        with self.assertRaises(ValueError):
            _ = Date(10000, 2, 1)

    def test_today(self):
        d = Date.today()
        self.assertIsInstance(d, Date)

    def test_today_with_tz(self):
        d = Date.today(tz=eastern)
        self.assertIsInstance(d, Date)

    def test_utc_today(self):
        d = Date.utc_today()
        self.assertIsInstance(d, Date)

    def test_from_timestamp_without_tz(self):
        d = Date.from_timestamp(0)
        self.assertEqual(d, Date(1970, 1, 1))

    def test_from_timestamp_with_tz(self):
        d = Date.from_timestamp(0, tz=eastern)
        self.assertEqual(d, Date(1969, 12, 31))

    def test_utc_from_timestamp(self):
        d = Date.utc_from_timestamp(0)
        self.assertEqual(d, Date(1970, 1, 1))

    def test_from_ordinal(self):
        d = Date.from_ordinal(1)
        self.assertEqual(d, Date(1, 1, 1))

    def test_parse(self):
        d = Date.parse("2018-04-30")
        self.assertEqual(d, Date(2018, 4, 30))

    def test_bad_parse_1(self):
        with self.assertRaises(ValueError):
            _ = Date.parse("30 April 2018")

    def test_bad_parse_2(self):
        with self.assertRaises(ValueError):
            _ = Date.parse("2018-04")

    def test_bad_parse_3(self):
        with self.assertRaises(ValueError):
            _ = Date.parse(object())

    def test_replace(self):
        d1 = Date(2018, 4, 30)
        d2 = d1.replace(year=2017)
        self.assertEqual(d2, Date(2017, 4, 30))

    def test_from_clock_time(self):
        d = Date.from_clock_time((0, 0), epoch=UnixEpoch)
        self.assertEqual(d, Date(1970, 1, 1))

    def test_bad_from_clock_time(self):
        with self.assertRaises(ValueError):
            _ = Date.from_clock_time(object(), None)

    def test_is_leap_year(self):
        self.assertTrue(Date.is_leap_year(2000))
        self.assertFalse(Date.is_leap_year(2001))

    def test_days_in_year(self):
        self.assertEqual(Date.days_in_year(2000), 366)
        self.assertEqual(Date.days_in_year(2001), 365)

    def test_days_in_month(self):
        self.assertEqual(Date.days_in_month(2000, 1), 31)
        self.assertEqual(Date.days_in_month(2000, 2), 29)
        self.assertEqual(Date.days_in_month(2001, 2), 28)

    def test_instance_attributes(self):
        d = Date(2018, 4, 30)
        self.assertEqual(d.year, 2018)
        self.assertEqual(d.month, 4)
        self.assertEqual(d.day, 30)
        self.assertEqual(d.year_month_day, (2018, 4, 30))
        self.assertEqual(d.year_week_day, (2018, 18, 1))
        self.assertEqual(d.year_day, (2018, 120))

    def test_can_add_years(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(years=2)
        self.assertEqual(d2, Date(1978, 6, 13))

    def test_can_add_negative_years(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(years=-2)
        self.assertEqual(d2, Date(1974, 6, 13))

    def test_can_add_years_and_months(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(years=2, months=3)
        self.assertEqual(d2, Date(1978, 9, 13))

    def test_can_add_negative_years_and_months(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(years=-2, months=-3)
        self.assertEqual(d2, Date(1974, 3, 13))

    def test_can_retain_offset_from_end_of_month(self):
        d = Date(1976, 1, -1)
        self.assertEqual(d, Date(1976, 1, 31))
        d += Duration(months=1)
        self.assertEqual(d, Date(1976, 2, 29))
        d += Duration(months=1)
        self.assertEqual(d, Date(1976, 3, 31))
        d += Duration(months=1)
        self.assertEqual(d, Date(1976, 4, 30))
        d += Duration(months=1)
        self.assertEqual(d, Date(1976, 5, 31))
        d += Duration(months=1)
        self.assertEqual(d, Date(1976, 6, 30))

    def test_can_roll_over_end_of_year(self):
        d = Date(1976, 12, 1)
        self.assertEqual(d, Date(1976, 12, 1))
        d += Duration(months=1)
        self.assertEqual(d, Date(1977, 1, 1))

    def test_can_add_months_and_days(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(months=1, days=1)
        self.assertEqual(d2, Date(1976, 7, 14))

    def test_can_add_months_then_days(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(months=1) + Duration(days=1)
        self.assertEqual(d2, Date(1976, 7, 14))

    def test_cannot_add_seconds(self):
        d1 = Date(1976, 6, 13)
        with self.assertRaises(ValueError):
            _ = d1 + Duration(seconds=1)

    def test_adding_empty_duration_returns_self(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration()
        self.assertIs(d1, d2)

    def test_adding_object(self):
        d1 = Date(1976, 6, 13)
        with self.assertRaises(TypeError):
            _ = d1 + object()

    def test_can_add_days_then_months(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(days=1) + Duration(months=1)
        self.assertEqual(d2, Date(1976, 7, 14))

    def test_can_add_months_and_days_for_last_day_of_short_month(self):
        d1 = Date(1976, 6, 30)
        d2 = d1 + Duration(months=1, days=1)
        self.assertEqual(d2, Date(1976, 8, 1))

    def test_can_add_months_then_days_for_last_day_of_short_month(self):
        d1 = Date(1976, 6, 30)
        d2 = d1 + Duration(months=1) + Duration(days=1)
        self.assertEqual(d2, Date(1976, 8, 1))

    def test_can_add_days_then_months_for_last_day_of_short_month(self):
        d1 = Date(1976, 6, 30)
        d2 = d1 + Duration(days=1) + Duration(months=1)
        self.assertEqual(d2, Date(1976, 8, 1))

    def test_can_add_months_and_days_for_last_day_of_long_month(self):
        d1 = Date(1976, 1, 31)
        d2 = d1 + Duration(months=1, days=1)
        self.assertEqual(d2, Date(1976, 3, 1))

    def test_can_add_months_then_days_for_last_day_of_long_month(self):
        d1 = Date(1976, 1, 31)
        d2 = d1 + Duration(months=1) + Duration(days=1)
        self.assertEqual(d2, Date(1976, 3, 1))

    def test_can_add_days_then_months_for_last_day_of_long_month(self):
        d1 = Date(1976, 1, 31)
        d2 = d1 + Duration(days=1) + Duration(months=1)
        self.assertEqual(d2, Date(1976, 3, 1))

    def test_can_add_negative_months_and_days(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(months=-1, days=-1)
        self.assertEqual(d2, Date(1976, 5, 12))

    def test_can_add_negative_months_then_days(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(months=-1) + Duration(days=-1)
        self.assertEqual(d2, Date(1976, 5, 12))

    def test_can_add_negative_days_then_months(self):
        d1 = Date(1976, 6, 13)
        d2 = d1 + Duration(days=-1) + Duration(months=-1)
        self.assertEqual(d2, Date(1976, 5, 12))

    def test_can_add_negative_months_and_days_for_first_day_of_month(self):
        d1 = Date(1976, 6, 1)
        d2 = d1 + Duration(months=-1, days=-1)
        self.assertEqual(d2, Date(1976, 4, 30))

    def test_can_add_negative_months_then_days_for_first_day_of_month(self):
        d1 = Date(1976, 6, 1)
        d2 = d1 + Duration(months=-1) + Duration(days=-1)
        self.assertEqual(d2, Date(1976, 4, 30))

    def test_can_add_negative_days_then_months_for_last_day_of_month(self):
        d1 = Date(1976, 6, 1)
        d2 = d1 + Duration(days=-1) + Duration(months=-1)
        self.assertEqual(d2, Date(1976, 4, 30))

    def test_can_add_negative_month_for_last_day_of_long_month(self):
        d1 = Date(1976, 5, 31)
        d2 = d1 + Duration(months=-1)
        self.assertEqual(d2, Date(1976, 4, 30))

    def test_can_add_negative_month_for_january(self):
        d1 = Date(1976, 1, 31)
        d2 = d1 + Duration(months=-1)
        self.assertEqual(d2, Date(1975, 12, 31))

    def test_subtract_date(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertEqual(new_year - christmas, Duration(days=7))

    def test_subtract_duration(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertEqual(new_year - Duration(days=7), christmas)

    def test_subtract_object(self):
        new_year = Date(2000, 1, 1)
        with self.assertRaises(TypeError):
            _ = new_year - object()

    def test_date_less_than(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertLess(christmas, new_year)

    def test_date_less_than_object(self):
        d = Date(2000, 1, 1)
        with self.assertRaises(TypeError):
            _ = d < object()

    def test_date_less_than_or_equal_to(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertLessEqual(christmas, new_year)

    def test_date_less_than_or_equal_to_object(self):
        d = Date(2000, 1, 1)
        with self.assertRaises(TypeError):
            _ = d <= object()

    def test_date_greater_than_or_equal_to(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertGreaterEqual(new_year, christmas)

    def test_date_greater_than_or_equal_to_object(self):
        d = Date(2000, 1, 1)
        with self.assertRaises(TypeError):
            _ = d >= object()

    def test_date_greater_than(self):
        new_year = Date(2000, 1, 1)
        christmas = Date(1999, 12, 25)
        self.assertGreater(new_year, christmas)

    def test_date_greater_than_object(self):
        d = Date(2000, 1, 1)
        with self.assertRaises(TypeError):
            _ = d > object()

    def test_date_equal(self):
        d1 = Date(2000, 1, 1)
        d2 = Date(2000, 1, 1)
        self.assertEqual(d1, d2)

    def test_date_not_equal(self):
        d1 = Date(2000, 1, 1)
        d2 = Date(2000, 1, 2)
        self.assertNotEqual(d1, d2)

    def test_date_not_equal_to_object(self):
        d1 = Date(2000, 1, 1)
        self.assertNotEqual(d1, object())

    def test_year_week_day(self):
        for ordinal in range(Date(2001, 1, 1).to_ordinal(), Date(2008, 1, 1).to_ordinal()):
            self.assertEqual(Date.from_ordinal(ordinal).iso_calendar(), date.fromordinal(ordinal).isocalendar())

    def test_time_tuple(self):
        d = Date(2018, 4, 30)
        self.assertEqual(d.time_tuple(), struct_time((2018, 4, 30, 0, 0, 0, 0, 120, -1)))

    def test_to_clock_time(self):
        d = Date(2018, 4, 30)
        self.assertEqual(d.to_clock_time(UnixEpoch), (1525046400, 0))
        self.assertEqual(d.to_clock_time(d), (0, 0))
        with self.assertRaises(TypeError):
            _ = d.to_clock_time(object())

    def test_weekday(self):
        d = Date(2018, 4, 30)
        self.assertEqual(d.weekday(), 0)

    def test_iso_weekday(self):
        d = Date(2018, 4, 30)
        self.assertEqual(d.iso_weekday(), 1)

    def test_str(self):
        self.assertEqual(str(Date(2018, 4, 30)), "2018-04-30")
        self.assertEqual(str(Date(0, 0, 0)), "0000-00-00")

    def test_repr(self):
        self.assertEqual(repr(Date(2018, 4, 30)), "neo4j.time.Date(2018, 4, 30)")
        self.assertEqual(repr(Date(0, 0, 0)), "neo4j.time.ZeroDate")

    def test_format(self):
        d = Date(2018, 4, 30)
        with self.assertRaises(NotImplementedError):
            _ = d.__format__("")

    def test_from_native(self):
        native = date(2018, 10, 1)
        d = Date.from_native(native)
        self.assertEqual(d.year, native.year)
        self.assertEqual(d.month, native.month)
        self.assertEqual(d.day, native.day)

    def test_to_native(self):
        d = Date(2018, 10, 1)
        native = d.to_native()
        self.assertEqual(d.year, native.year)
        self.assertEqual(d.month, native.month)
        self.assertEqual(d.day, native.day)

    def test_iso_format(self):
        d = Date(2018, 10, 1)
        self.assertEqual("2018-10-01", d.iso_format())

    def test_from_iso_format(self):
        expected = Date(2018, 10, 1)
        actual = Date.from_iso_format("2018-10-01")
        self.assertEqual(expected, actual)
